package jumpstart.min.business.commons.interpreter;

import java.io.Serializable;
import java.sql.BatchUpdateException;
import java.sql.SQLException;

import jumpstart.min.business.commons.exception.CannotDeleteIsReferencedException;
import jumpstart.min.business.commons.exception.DuplicateAlternateKeyException;
import jumpstart.min.business.commons.exception.DuplicatePrimaryKeyException;
import jumpstart.min.business.commons.exception.ServiceException;
import jumpstart.min.business.commons.exception.UnexpectedException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class JBossExceptionInterpreter implements IEJBExceptionInterpreter {
	static private final Log LOG = LogFactory.getLog(JBossExceptionInterpreter.class);

	static public final int ERRORCODE_HSQLDB_INTEGRITY_CONSTRAINT_VIOLATION = -8;
	static public final int ERRORCODE_HSQLDB_UNIQUE_CONSTRAINT_VIOLATION = -104;
	static public final int ERRORCODE_MYSQL_CANNOT_DELETE_OR_UPDATE_A_PARENT_ROW = 1451;
	static public final int ERRORCODE_MYSQL_CANNOT_DELETE_OR_UPDATE_A_CHILD_ROW = 1452;
	static public final int ERRORCODE_MYSQL_DUPLICATE_ENTRY = 1062;
	static public final int ERRORCODE_UNKNOWN = 0;

	public ServiceException interpret(Exception e) {

		// Assume it's one of JBoss/Hibernate's deplorable deeply wrapped
		// exceptions - use its utility to find the root cause

		Throwable rootCause = org.hibernate.exception.ExceptionUtils.getRootCause(e);

		if (rootCause instanceof ServiceException) {
			// It's one of ours! But somehow it got wrapped up, maybe by an
			// Entity Lifecycle callback?
			return (ServiceException) rootCause;
		}
		else if (rootCause instanceof org.hibernate.NonUniqueObjectException) {
			org.hibernate.NonUniqueObjectException rootE = (org.hibernate.NonUniqueObjectException) rootCause;
			//String technicalMessage = rootE.getLocalizedMessage();
			//String entityName = rootE.getEntityName();
			Serializable identifier = rootE.getIdentifier();
			return new DuplicatePrimaryKeyException(identifier);
		}
		else if (rootCause instanceof SQLException) {
			SQLException rootE = (SQLException) rootCause;
			String rootMsg = rootE.getMessage();
			int errorCode = rootE.getErrorCode();

			if (errorCode == ERRORCODE_MYSQL_DUPLICATE_ENTRY) {
				// Expect message starting with "Duplicate entry"
				String technicalMessage = rootE.getLocalizedMessage();
				String[] chunks = rootMsg.split("\'");
				if (rootMsg.indexOf("for key 1") > -1) {
					Serializable identifier = chunks[1];
					return new DuplicatePrimaryKeyException(identifier);
				}
				else {
					return new DuplicateAlternateKeyException(technicalMessage);
				}
			}
			else if (errorCode == ERRORCODE_HSQLDB_UNIQUE_CONSTRAINT_VIOLATION) {
				String[] chunks = rootE.getMessage().split(" ");
				if (chunks[2].equals("violation:") && chunks[4].equals("in")) {
					String constraintName = chunks[3];
					String technicalMessage = rootE.getLocalizedMessage();
					if (constraintName.startsWith("SYS_CT")) {
						return new DuplicateAlternateKeyException(technicalMessage);
					}
				}
			}
			else if (errorCode == ERRORCODE_HSQLDB_INTEGRITY_CONSTRAINT_VIOLATION) {
				String[] chunks = rootE.getMessage().split(" ");
				if (chunks[4].equals("table:") && chunks[6].equals("in")) {
					String entityName = chunks[5];
					//String constraintName = chunks[3];
					return new CannotDeleteIsReferencedException(entityName);
				}
			}
			else if (errorCode == ERRORCODE_MYSQL_CANNOT_DELETE_OR_UPDATE_A_CHILD_ROW
					|| errorCode == ERRORCODE_MYSQL_CANNOT_DELETE_OR_UPDATE_A_PARENT_ROW) {
				// else if (rootMsg.indexOf("a foreign key constraint
				// fails") > -1) {
				//String technicalMessage = rootE.getLocalizedMessage();
				// Extract the entity name and property name with regex.
				// An example of the message is:
				// Cannot delete or update a parent row: a foreign key
				// constraint fails (`perspectives/code_group`,
				// CONSTRAINT
				// `FKDCDB21ADA2BDAF42` FOREIGN KEY (`status_code`)
				// REFERENCES `code` (`id`))
				String[] chunks = rootE.getMessage().split("[`/]");
				if (chunks[3].equals(", CONSTRAINT ") && chunks[5].equals(" FOREIGN KEY (")) {
					String referencedByEntityName = chunks[2];
					//String constraintName = chunks[4];
					String referencedByPropertyName = chunks[6];
					return new CannotDeleteIsReferencedException(referencedByEntityName, referencedByPropertyName);
				}
			}
			else if (errorCode == ERRORCODE_UNKNOWN) {
				if (rootE instanceof BatchUpdateException) {
					// Have seen this in HSQLDB, and it was due to duplicate primary key!!
					return new DuplicatePrimaryKeyException();
				}
			}
		}
		else if (rootCause instanceof org.hibernate.PropertyValueException) {
			org.hibernate.PropertyValueException rootE = (org.hibernate.PropertyValueException) rootCause;
			String rootMsg = rootE.getMessage();
			//String technicalMessage = rootE.getLocalizedMessage();
			if (rootMsg.indexOf("not-null property references") > -1) {
				String[] qualifiedEntityNameParts = rootE.getEntityName().split("\\.");
				String referencedByEntityName = qualifiedEntityNameParts[qualifiedEntityNameParts.length - 1];
				String referencedByPropertyName = rootE.getPropertyName();
				return new CannotDeleteIsReferencedException(referencedByEntityName, referencedByPropertyName);
			}
		}
		// else if (e instanceof EntityExistsException) {
		// Throwable cause = e.getCause();
		//
		// if (cause instanceof org.hibernate.NonUniqueObjectException) {
		// findings = DUP_KEY;
		// _interpretedException = cause;
		// }
		// else if (cause instanceof
		// org.hibernate.exception.ConstraintViolationException) {
		// findings = OTHER_CONSTRAINT_VIOLATION;
		// ConstraintViolationException violation =
		// (ConstraintViolationException) cause;
		// Throwable cause1 = violation.getCause();
		// if (cause1 instanceof SQLException) {
		// SQLException sqlE = (SQLException) cause1;
		// if (sqlE.getMessage().indexOf("Duplicate entry") > -1) {
		// findings = DUP_ALT_KEY;
		// }
		// else {
		// findings = OTHER_CONSTRAINT_VIOLATION;
		// }
		// _interpretedException = sqlE;
		// _technicalMessage = sqlE.getLocalizedMessage();
		// }
		// else if (cause1 instanceof BatchUpdateException) {
		// BatchUpdateException batchUpdateE = (BatchUpdateException) cause1;
		// if (batchUpdateE.getMessage().indexOf("Duplicate entry") > -1) {
		// findings = DUP_ALT_KEY;
		// }
		// else {
		// findings = OTHER_CONSTRAINT_VIOLATION;
		// }
		// _interpretedException = batchUpdateE;
		// _technicalMessage = batchUpdateE.getLocalizedMessage();
		// }
		// }
		// }
		// else if (e instanceof PersistenceException) {
		// Throwable cause = e.getCause();
		//
		// if (cause instanceof org.hibernate.NonUniqueObjectException) {
		// findings = DUP_KEY;
		// _interpretedException = cause;
		// }
		// else if (cause instanceof
		// org.hibernate.exception.GenericJDBCException) {
		// org.hibernate.exception.GenericJDBCException causeGeneric =
		// (org.hibernate.exception.GenericJDBCException) cause;
		// Throwable nested = causeGeneric.getCause();
		// if (nested instanceof java.sql.BatchUpdateException) {
		// // In JBoss 4.0.4.CR2, this is probably caused by duplicate alternate
		// key,
		// // or even non-unique primary key, but there's NO info to help us
		// know.
		// findings = INSCRUTABLE_BATCH_UPDATE_EXCEPTION;
		// _interpretedException = nested;
		// }
		// }
		// }
		// else if (e instanceof
		// org.hibernate.exception.ConstraintViolationException) {
		// findings = OTHER_CONSTRAINT_VIOLATION;
		// ConstraintViolationException violation =
		// (ConstraintViolationException) e;
		// Throwable cause1 = violation.getCause();
		// if (cause1 instanceof SQLException) {
		// SQLException sqlE = (SQLException) cause1;
		// if (_technicalMessage.indexOf("Duplicate entry") > -1) {
		// findings = DUP_ALT_KEY;
		// }
		// else {
		// findings = OTHER_CONSTRAINT_VIOLATION;
		// }
		// _interpretedException = sqlE;
		// _technicalMessage = sqlE.getLocalizedMessage();
		// }
		// }
		// else {
		// Throwable cause = e.getCause();
		// if (cause instanceof InvocationTargetException) {
		// InvocationTargetException invoc = (InvocationTargetException) cause;
		// Throwable targetE = invoc.getTargetException();
		// if (targetE instanceof BusinessException) {
		// // No worries it's one of ours! But somehow it got wrapped up, maybe
		// by an Entity Lifecycle callback?
		// findings = BUSINESS_EXCEPTION;
		// _interpretedException = e;
		// _message = e.getLocalizedMessage();
		// }
		// }
		// }

		LOG.error("Unexpected exception.  Root cause follows...");
		LOG.error(rootCause);

		return new UnexpectedException(rootCause);
	}

}
